from typing import Any, Literal

from .base import BaseEvent


class ChunkEvent(BaseEvent):
    """Event emitted when a chunk of data is received."""
    type: Literal["CHUNK"] = "CHUNK"

    chunk: Any
    """The chunk of output generated by the step generator."""
